from .apps import DesktopApp, Website, quick_website, quick_desktop_app
from .pages import Page
from .elements import Element, IFrameElement
from .structure import ToUIBlueprint

__version__ = "v1.0.3-beta"



"""
A module for containing classes that help in organizing the apps.
"""

import flask
from copy import copy


class ToUIBlueprint(flask.Blueprint):
    """
    A class that inherits `flask.Blueprint`.

    The difference between this class and `flask.Blueprint` is that you can add `Page`
    objects to this class. To learn how to use blueprints, check `flask's` documentation
    https://flask.palletsprojects.com/.

    Examples
    --------

    Creating a blueprint:

    >>> from toui import ToUIBlueprint
    >>> blueprint = ToUIBlueprint("Blueprint", __name__)

    Adding a `Page` to a blueprint:

    >>> from toui import Page
    >>> page = Page(html_str="<h1>This page is part of a blueprint</h1>", url="/")
    >>> blueprint.add_pages(page)

    Creating a `Website` and adding the blueprint to it:

    >>> from toui import Website
    >>> app = Website(__name__)
    >>> app.register_toui_blueprint(blueprint)

    See Also
    --------
    flask.Blueprint

    """
    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        args
            `flask.Blueprint` arguments.
        kwargs
            `flask.Blueprint` keyword arguments.
        """
        super(ToUIBlueprint, self).__init__(*args, **kwargs)
        self.pages = []

    def add_pages(self, *pages, do_copy=False):
        """Adds pages to the blueprint.

        Parameters
        ----------
        pages: list(Page)
            List of 'Page' objects.

        """
        for page in pages:
            if do_copy:
                page = copy(page)
            self.pages.append(page)

    def register_toui_blueprint(self, blueprint, **options):
        """
        Registers a `ToUIBlueprint` object. It is similar to `flask.Blueprint.register_blueprint`.

        Parameters
        ----------
        blueprint: toui.structure.ToUIBlueprint

        options
            Same as `flask.Blueprint.register_blueprint` `options` parameter.

        See Also
        --------
        toui.structure.ToUIBlueprint
        flask.Blueprint.register_blueprint

        """
        self.add_pages(*blueprint.pages)
        self.register_blueprint(blueprint, **options)


if __name__ == "__main__":
    import doctest
    doctest.testmod()


"""
Dynamic web page
"""
from toui import Page, Website


# create website
app = Website(__name__, assets_folder="assets", secret_key="some text")

# create page
main_pg = Page(html_file="assets/test3.html", url="/")

# create a function and add it to the page
def addValues():
    page = app.get_user_page()
    input1 = page.get_element("input-1").get_value()
    input2 = page.get_element("input-2").get_value()
    result = float(input1) + float(input2)
    page.get_element('output').set_value(result)

add_button = main_pg.get_element('add-button') # 'add-button' is an id of a <button>
add_button.onclick(addValues)

# add the page to the website
app.add_pages(main_pg)

# run
if __name__ == "__main__":
    app.run(debug=True)


"""
A module that creates web apps and desktop apps.
"""
import __main__
from flask import Flask, session
from flask_sock import Sock
import webview
import json
import inspect
import time
import os
from copy import copy
from abc import ABCMeta, abstractmethod
from flask_basicauth import BasicAuth
from toui._helpers import warn, info, debug, error
from toui.pages import Page
from toui.exceptions import ToUIWrongPlaceException
from toui._defaults import validate_ws, validate_data


class _App(metaclass=ABCMeta):
    """The base class for DesktopApp and Website"""

    def _add_function(self, func):
        name = func.__name__
        if not callable(func):
            warn(f"Variable '{name}' is not a function.")
            return
        if name.startswith("_"):
            warn(f"Function '{name}' starts with '_'. It is safer to avoid functions that starts with '_'"
                 f"because they might overlap with functions used by the package.")
        if self._func_exists(name):
            warn(f"Function '{name}' exists.")
        self._functions[name] = func

    @abstractmethod
    def add_pages(self): pass

    @abstractmethod
    def open_new_page(self, url): pass

    @abstractmethod
    def get_user_page(self): pass

    @abstractmethod
    def user_vars(self): pass

    @abstractmethod
    def _communicate(self): pass

    @abstractmethod
    def run(self): pass

    def _add_functions(self, *functions):
        for func in functions:
            self._add_function(func)

    def _get_functions(self):
        """Gets all added functions in this class. This is a private function."""
        return self._functions

    def _func_exists(self, func_name: str):
        """Checks if a function exists. This is a private function."""
        if func_name in self._get_functions().keys():
            return True
        else:
            return False

    def _call_func(self, func_name, *args, page=None):
        """Calls a function in this class. Its return value depends on the function called. This is a private function."""
        functions = self._get_functions()
        info(f'"{func_name}" called')
        if page:
            func = _FuncWithPage(functions[func_name], page)
            return func(*args)
        return functions[func_name](*args)


class _FuncWithPage:
    def __init__(self, func, page):
        self.func = func
        self.page = page
    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

        
class Website(_App):
    """
    A class that create a web application from HTML files.

    Attributes
    ----------
    flask_app: Flask
        ToUI creates web applications using `Flask`. You can access the `Flask` object using the attribute `flask_app`.

    forbidden_urls: list
        These are URLs that ToUI does not allow the user to use because ToUI uses them.

    user_vars: dict
        A dictionary that stores data unique to each user. The data are stored in `flask` `session` object.

    pages: list
        A list of added `Page` objects.

    Examples
    --------

    Creating a web app:

    >>> from toui import Website
    >>> app = Website(__name__, secret_key="some key")

    Creating a page and adding it to the app:

    >>> from toui import Page
    >>> home_page = Page(html_str="<h1>This is the home page</h1>")
    >>> app.add_pages(home_page)

    Running the app:

    >>> if __name__ == "__main__":
    ...     app.run(debug=True) # doctest: +SKIP

    See Also
    --------
    :py:class:`toui.pages.Page`
    :py:class:`DesktopApp`

    """

    _auth = None
    _default_vars = {}
    
    def __init__(self, name=None, assets_folder=None, secret_key=None):
        """

        Parameters
        ----------
        name: str (optional)
            The name of the app.

        assets_folder: str (optional)
            The path to the folder that contains the HTML files and other assets.

        secret_key: str (optional)
            Sets the `secret_key` attribute for `flask.Flask`

        """
        self._functions = {}
        if not name:
            if hasattr(__main__, "__file__"):
                name = os.path.basename(__main__.__file__).split(".")[0]
            else:
                name = "app"
        if not assets_folder:
            assets_folder = "/"
        self.flask_app = Flask(name, static_folder=assets_folder, static_url_path="/")
        if secret_key is None:
            warn("No secret key was set. Generating a random secret key for Flask.")
            secret_key = os.urandom(50)
        self.flask_app.secret_key = secret_key
        self._socket = Sock(self.flask_app)
        self.pages = []
        self._socket.route("/toui-communicate")(self._communicate)
        self.forbidden_urls = ['/toui-communicate']
        self._validate_ws = validate_ws
        self._validate_data = validate_data

    def add_pages(self, *pages, do_copy=False, blueprint=None, endpoint=None):
        """
        Adds pages to the app.

        Parameters
        ----------
        pages: list(Page)
            List of `Page` objects.

        do_copy: bool, default = False
            If ``True``, the `Page` will be copied before adding to the app.

        blueprint: toui.structure.ToUIBlueprint, flask.Blueprint, default = None
            If a `flask.Blueprint` or a `ToUIBlueprint` was added, the `Page` view_func will be added
            to the blueprint instead of the `Flask` app.

        endpoint
            `endpoint` parameter in `flask.Flask.route`.
            
        """
        for page in pages:
            if page.url in self.forbidden_urls:
                warn(f"The URL '{page.url}' is not allowed and might cause errors.")
            if do_copy:
                page = copy(page)
            page._add_script()
            page._app = self
            self.pages.append(page)
            view_func = page.view_func
            if self._auth:
                view_func = self._auth.required(view_func)
            if not endpoint:
                endpoint = page.url
            if blueprint:
                route = blueprint.route(page.url, methods=['GET', 'POST'], endpoint=page.url)(view_func)
            else:
                route = self.flask_app.route(page.url, methods=['GET', 'POST'], endpoint=page.url)(view_func)
            for func in page._functions.values():
                self._add_functions(func)

    def open_new_page(self, url):
        """
        Redirects to another URL.

        This function should only be called after the app starts running.

        Parameters
        ----------
        url: str
            URL of the new page.

        Returns
        -------
        None

        """
        try:
            session.keys()
            self.get_user_page()._open_another_page(url)
        except RuntimeError:
            raise ToUIWrongPlaceException(f"The function `{inspect.currentframe().f_code.co_name}` should only be called after the app runs.")

    def get_user_page(self):
        """
        Returns the current `Page`.

        This function should only be called after the app starts running.

        Returns
        -------
        pg: Page

        """
        try:
            pg = session['user page']
            return pg
        except RuntimeError as e:
            raise ToUIWrongPlaceException(f"The function `{inspect.currentframe().f_code.co_name}` should only be called after the app runs.")

    def run(self, *args, **kwargs):
        """
        Runs the app. It calls the function `flask.Flask.run`. The arguments will be passed to
        `flask.Flask.run` function.

        Parameters
        ----------
        args: Any

        kwargs: Any

        """
        self.flask_app.run(*args, **kwargs)

    @property
    def user_vars(self):
        session_exists = self._session_check()
        if session_exists:
            return session['variables']
        else:
            return self._default_vars

    def _session_check(self):
        """This is a private function."""
        try:
            session.keys()
        except RuntimeError as e:
            return False
        if not "variables" in session.keys():
            session['variables'] = self._default_vars
        if not "user page" in session.keys():
            session['user page'] = None
        return True

    def _communicate(self, ws):
        """This is a private function."""
        validation = self._validate_ws(ws)
        if not validation:
            info("WebSocket validation returns `False`. No data should be sent or received.")
            return
        info(f'WebSocket connected: {ws.connected}')
        while True:
            data_from_js = ws.receive()
            data_validation = self._validate_data(data_from_js)
            if not data_validation:
                info("Data validation returns `False`. The data will not be used.")
                continue
            debug(data_from_js)
            s = time.time()
            self._session_check()
            data_dict = json.loads(data_from_js)
            func = data_dict['func']
            args = data_dict['args']
            url = data_dict['url']
            new_html = data_dict['html']
            new_page = Page(url=url)
            new_page.from_str(new_html)
            new_page._app = self
            new_page._signal_mode = True
            new_page._ws = ws
            session['user page'] = new_page
            if self._func_exists(func):
                self._call_func(func, *args)
            del session['user page']
            e = time.time()
            debug(f"TIME: {e - s}s")

    # Website-specific methods
    def set_restriction(self, username, password):
        """
        Makes the app private.

        Adds a username and password to the app.

        Parameters
        ----------
        username: str

        password: str

        """
        self._auth = BasicAuth(self.flask_app)
        self.flask_app.config['BASIC_AUTH_USERNAME'] = username
        self.flask_app.config['BASIC_AUTH_PASSWORD'] = password

    def set_ws_validation(self, func):
        """
        Validate `simple_websocket.ws.Server` object before sending and accepting data.

        ToUI uses Flask-Sock for websocket communication. Flask-Sock generates a
        `simple_websocket.ws.Server` object when a connection is established. If you
        wanted to access this object before sending and receiving data, input a function
        that has one argument `ws`. This function should either return ``True`` or ``False``.
        If the function returns ``False``, no data will be sent or received using ToUI with
        the client.

        Parameters
        ----------
        func: Callable
            A function that validates the Server object. It should have one argument `ws` and
            should either return ``True`` or ``False``.

        See Also
        --------
        flask_sock
        simple_websockets

        """
        self._validate_ws = func


    def set_data_validation(self, func):
        """
        Validate data received from JavaScript before using it.

        ToUI receives data from JavaScript in the form of JSON object. To validate this data
        before allowing ToUI to use it, input a function that checks the data. This function
        should have one argument `data` and should either return ``True`` or ``False``.
        If the function returns ``False``, the data will not be used by ToUI.

        Currently, the JSON object contains the following keys:

        {func: ...,
         args: ...,
         url: ...,
         html: ...}

        `func` contains the name of the Python function that should be called, `args` are the
        arguments of this function, `url` is the URL of the HTML page that sent the data, 'html'
        if the HTML document itself as a string.
        However, note that the structure of the JSON object might change in future versions of
        ToUI.

        Parameters
        ----------
        func: Callable
            A function that validates data received from JavaScript. It should have one argument
            `data` and should either return ``True`` or ``False``.

        See Also
        --------
        Website.set_ws_validation

        """
        self._validate_data = func


    def register_toui_blueprint(self, blueprint, **options):
        """
        Registers a `ToUIBlueprint` object. It is similar to `Flask.register_blueprint`.

        Parameters
        ----------
        blueprint: toui.structure.ToUIBlueprint

        options
            Same as `Flask.register_blueprint` `options` parameter.

        See Also
        --------
        toui.structure.ToUIBlueprint
        flask.Blueprint

        """
        self.add_pages(*blueprint.pages, blueprint=blueprint)
        self.flask_app.register_blueprint(blueprint=blueprint, **options)


class _DesktopAppApi:

    def __init__(self, app):
        self.app = app

    def communicate(self, data_from_js):
        self.app._communicate(data_from_js, self.window)


class DesktopApp(_App):
    """
    A class that create a desktop application from HTML files.

    Attributes
    ----------
    name: str
        The name of the app

    pages: list
        A list of added `Page` objects.

    user_vars: dict
        A normal dictionary. It was created only to make the API for the `Website` class and
        `DesktopApp` class similar.

    Examples
    --------

    Creating a desktop app:

    >>> from toui import DesktopApp
    >>> app = DesktopApp("MyApp")

    Creating a page and adding it to the app:

    >>> from toui import Page
    >>> home_page = Page(html_str="<h1>This is the home page</h1>")
    >>> app.add_pages(home_page)

    Running the app:

    >>> if __name__ == "__main__":
    ...     app.run() # doctest: +SKIP

    See Also
    --------
    :py:class:`toui.pages.Page`
    :py:class:`Website`

    """
    def __init__(self, name="App", assets_folder=None):
        """

        Parameters
        ----------
        name: str
            The name of the app.

        """
        self.name = name
        self.pages = []
        if not assets_folder:
            assets_folder = "."
        self._assets_folder = assets_folder
        self._variables = {}
        self._functions = {}
        self._is_running = False

    def add_pages(self, *pages, do_copy=False):
        """
        Adds pages to the app.

        Parameters
        ----------
        pages: list(Page)
            List of `Page` objects.

        do_copy: bool, default = False
            If ``True``, the `Page` will be copied before adding to the app.

        """
        for page in pages:
            if do_copy:
                page = copy(page)
            page._add_script("desktop")
            page._app = self
            self.pages.append(page)
            for func in page._functions.values():
                self._add_functions(func)

    def open_new_page(self, url):
        """
        Opens a new window that has the specified URL.

        This function should only be called after the app starts running.

        Parameters
        ----------
        url: str
            URL of the new page.

        Returns
        -------
        None

        """
        for p in self.pages:
            if p.url == url:
                api = _DesktopAppApi(self)
                p._create_window(name=self.name, api=api, assets_folder=self._assets_folder)

    def get_user_page(self):
        """
        Returns the current `Page`.

        This function should only be called after the app starts running.

        Returns
        -------
        pg: Page

        """
        if not self._is_running:
            raise ToUIWrongPlaceException(f"The function `get_user_page` should only be called after the app runs.")
        func_frame = inspect.currentframe()
        func_type = type(None)
        while func_type is not _FuncWithPage:
            func_frame = func_frame.f_back
            if "self" in func_frame.f_locals:
                func_type = func_frame.f_locals['self'].__class__
            if func_frame is None:
                raise Exception("Could not find page. Perhaps `get_user_page` was not called within a function"
                                "called by HTML event?")
        page = func_frame.f_locals['self'].page
        return page

    @property
    def user_vars(self):
        return self._variables

    def run(self, *args, **kwargs):
        """
        Runs the app. It calls the function `webview.start`. The arguments will be passed to
        `webview.start` function.

        Parameters
        ----------
        args: Any

        kwargs: Any

        """
        if len(self.pages) == 0:
            raise Exception("You need to have at least one `Page` added before running the app.")
        api = _DesktopAppApi(self)
        func = self.pages[0]._create_first_window(name=self.name, api=api, assets_folder=self._assets_folder)
        self._is_running = True
        webview.start(func)

    def _communicate(self, data_from_js, window):
        debug("communicating")
        s = time.time()
        data_dict = json.loads(data_from_js)
        debug(data_dict)
        func = data_dict['func']
        args = data_dict['args']
        new_html = data_dict['html']
        new_page = Page()
        new_page.from_str(new_html)
        new_page._app = self
        new_page._signal_mode = True
        new_page.window = window
        if self._func_exists(func):
            self._call_func(func, *args, page=new_page)
        e = time.time()
        debug(f"TIME: {e - s}s")


def quick_website(name="App", html_file=None, html_str=None, url="/", assets_folder=None, secret_key=None):
    """
    Creates a web app and adds a single `Page` to it.

    Parameters
    ----------
    name: str (optional)
        The name of the app.

    html_file: str (optional)
        The path to the HTML file that will be used to create the `Page`.

    html_str: str (optional)
        The content of the `Page`.

    url: str (optional)
        The URL of the `Page`.

    assets_folder: str (optional)
        The path to the folder that contains the HTML file. If no HTML files are used,
        you can ignore this parameter.

    secret_key: str (optional)
        Sets the `secret_key` attribute for `flask.Flask`

    Returns
    -------
    Website

    """
    app = Website(name=name, assets_folder=assets_folder, secret_key=secret_key)
    page = Page(url=url, html_file=html_file, html_str=html_str)
    app.add_pages(page)
    return app


def quick_desktop_app(name="App", html_file=None, html_str=None, url="/", assets_folder=None):
    """
    Creates a desktop app and adds a single `Page` to it.

    Parameters
    ----------
    name: str (optional)
        The name of the app.

    html_file: str (optional)
        The path to the HTML file that will be used to create the `Page`.

    html_str: str (optional)
        The content of the `Page`.

    url: str (optional)
        The URL of the `Page`.

    assets_folder: str (optional)
        The path to the folder that contains the HTML file. If no HTML files are used,
        you can ignore this parameter.

    Returns
    -------
    DesktopApp

    """
    app = DesktopApp(name=name, assets_folder=assets_folder)
    page = Page(url=url, html_file=html_file, html_str=html_str)
    app.add_pages(page)
    return app

if __name__ == "__main__":
    import doctest
    results = doctest.testmod()
    print(results)


"""
A module that creates web pages or windows.
"""
import os
import time
from bs4 import BeautifulSoup
import webview
import json
from toui.elements import Element
from toui._javascript_templates import custom_func, get_script
from copy import copy
from toui._helpers import warn, info, debug, selector_to_str
from toui._signals import Signal
from toui._defaults import view_func


class _PageSignal(Signal):
    """
    Creates signals that will be sent to JavaScript.

    These signals are related the methods of the `Page` object.
    """

    no_return_functions = ["add_function"]

    @staticmethod
    def from_str(**kwargs):
        js_func = "_setDoc"
        js_args = []
        js_kwargs = {"doc": kwargs['html_str']}
        return {'func': js_func, 'args': js_args, 'kwargs': js_kwargs}

    @staticmethod
    def from_bs4_soup(**kwargs):
        js_func = "_setDoc"
        js_args = []
        js_kwargs = {"doc": str(kwargs['bs4_soup'])}
        return {'func': js_func, 'args': js_args, 'kwargs': js_kwargs}

    @staticmethod
    def add_function(**kwargs):
        value = kwargs['return_value']
        if value != "":
            value = custom_func(value)
        js_func = "_addScript"
        js_args = []
        js_kwargs = {"script": value}
        return {'func': js_func, 'args': js_args, 'kwargs': js_kwargs}

    @staticmethod
    def _open_another_page(**kwargs):
        js_func = "_goTo"
        js_args = []
        js_kwargs = {"url": kwargs['url']}
        return {'func': js_func, 'args': js_args, 'kwargs': js_kwargs}


class Page:
    """
    Creates an HTML page that can be used as a window or as a webpage.

    Attributes
    ----------
    url: str
        The URL of the page.

    window: pywebview.Window, default = None
        A `pywebview.Window` object. It is automatically created when adding the `Page` to
        `DesktopApp` and running the app.

    view_func: Callable
        The view function of the `Page`. This is the function that will be decorated by
        `Flask.route()` in when creating web apps.

    Examples
    --------

    Importing the class:

    >>> from toui import Page

    Creating a `Page` from an HTML file:

    >>> path = "../examples/assets/test1.html"
    >>> page = Page(html_file=path)

    Creating a `Page` from a string:

    >>> page = Page(html_str="<html><h1>Hello</h1></html>")
    >>> page
    <html><h1>Hello</h1></html>

    Creating a page with a URL:

    >>> page = Page(html_str="<html><h1>Hello</h1></html>", url="/")

    Getting an element in the `Page` from its id:

    >>> page = Page(html_str='<html><h1 id="heading">Hello</h1></html>', url="/")
    >>> element = page.get_element(element_id="heading")
    >>> element
    <h1 id="heading">Hello</h1>

    See Also
    --------
    `pywebview api`: https://pywebview.flowrl.com/guide/api.html

    """

    def __init__(self, html_file=None, html_str=None, url=None):
        """
        Parameters
        ----------
        html_file: str
            The path to the HTML file.

        html_str: str
            A string containing HTML code.

        url: str, optional
            If the page was used as a webpage, this will be the URL of the page.

        """
        if html_file:
            self._html_file = html_file
            with open(html_file, "rt") as file:
                original_html = file.read()
            if url is None:
                url = "/" + os.path.basename(html_file)
        elif html_str:
            original_html = html_str
            self._html_file = None
        else:
            original_html = "<html></html>"
            self._html_file = None
        self._html = BeautifulSoup(original_html, features="html.parser")
        if self._html.find("html") is None:
            self._html = BeautifulSoup(f"<html>{original_html}</html>", features="html.parser")
        if url is None:
            url = "/"
        self.url = url

        # Other attributes
        self.window = None

        # Other internal attributes
        self._app = None
        self._signal_mode = False
        self._signals = []
        self._functions = {}
        self._view_func = lambda: view_func(self)

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return self.to_str()

    def __copy__(self):
        new_pg = Page(html_file=self._html_file, url=self.url)
        new_pg.from_bs4_soup(self.to_bs4_soup())
        new_pg.window = copy(self.window)
        new_pg._signal_mode = self._signal_mode
        new_pg._app = self._app
        return new_pg
    
    @_PageSignal()
    def from_str(self, html_str):
        """
        Converts HTML code to a `Page` object.

        Parameters
        ----------
        html_str: str
            HTML code.

        """
        self._html = BeautifulSoup(html_str, features="html.parser")

    def to_str(self):
        """
        Converts the `Page` object to HTML code.

        Returns
        -------
        str

        """
        return str(self._html)

    def from_html_file(self, html_path):
        """
        Reads an HTML file and converts it to an `Page` object.

        Parameters
        ----------
        html_path: str

        """
        with open(html_path, "rt") as file:
            self.from_str(file.read())

    def to_html_file(self, html_path):
        """
        Converts the `Page` object to an HTML file.

        Parameters
        ----------
        html_path: str

        """
        with open(html_path, "w") as file:
            file.write(self.to_str())

    @_PageSignal()
    def from_bs4_soup(self, bs4_soup):
        """
        Converts a `bs4.BeautifulSoap` object to a `Page` object.
        The `bs4.BeautifulSoap` object will be copied before converting.

        Parameters
        ----------
        bs4_soup: bs4.BeautifulSoap

        See Also
        --------
        bs4

        """
        self._html = copy(bs4_soup)

    def to_bs4_soup(self):
        """
        Converts the `Page` object to a `bs4.BeautifulSoap` object.
        The `Page` object will be copied before converting.

        Returns
        -------
        bs4.BeautifulSoap

        See Also
        --------
        bs4: https://beautiful-soup-4.readthedocs.io/en/latest/

        """
        return copy(self._html)

    def get_element(self, element_id, do_copy=False):
        """
        Gets an element from its ``id`` attribute. You can imagine this function as ``document.getElementById``
        in JavaScript.

        Creating a page and getting an element by its ``id``:

        >>> page = Page(html_str='<html><h1 id="heading">Hello</h1></html>', url="/")
        >>> element = page.get_element(element_id="heading")
        >>> element
        <h1 id="heading">Hello</h1>

        Parameters
        ----------
        element_id: str

        do_copy: bool, default = False
            If ``True``, the element will be copied.

        Returns
        -------
        element: Element
            If the element was found, an `Element` object will be returned.

        None
            If the element was not found.

        """
        bs4_tag = self._html.find(id=element_id)
        if bs4_tag is None:
            return None
        element = Element()
        if do_copy:
            element.from_bs4_tag(bs4_tag)
        else:
            element._from_bs4_tag_no_copy(bs4_tag)
        element._parent_page = self
        element._signal_mode = self._signal_mode
        element._selector = {"selector": f"[id={element_id}]"}
        return element

    def get_elements(self, tag_name=None, class_name=None, name=None, do_copy=False, attrs=None):
        """
        Get elements from the `Page` by their tag name and attributes.

        Parameters
        ----------
        tag_name: str, default=None
            The tag name of the elements.

        class_name: str, default=None
            The value of the ``class`` attribute of the elements.

        name: str, default=None
            The value of the ``name`` attribute of the elements.

        do_copy: bool
            If ``True``, the elements will be copied.

        attrs: dict, default=None
            Attributes other than ``class`` and ``name`` can be specified in this dictionary.

        Returns
        -------
        elements_list: List
            A list of `Element` objects that match the parameters.

        """
        if attrs is None:
            attrs = {}
        if class_name:
            attrs['class'] = class_name
        if name:
            attrs['name'] = name
        bs4_tags = self._html.find_all(name=tag_name, attrs=attrs)
        elements_list = []
        for tag_num, bs4_tag in enumerate(bs4_tags):
            element = Element()
            if do_copy:
                element.from_bs4_tag(bs4_tag)
            else:
                element._from_bs4_tag_no_copy(bs4_tag)
            element._parent_page = self
            element._signal_mode = self._signal_mode
            element._selector = {"selector": selector_to_str(tag_name=tag_name, class_name=class_name,
                                                             name=name, attrs=attrs),
                                 "number": tag_num}
            elements_list.append(element)
        return elements_list

    def get_html_element(self):
        """
        Gets the first ``<html>`` element.

        Returns
        -------
        Element

        None
            If the ``<html>`` element was not found.

        """
        elements = self.get_elements("html")
        if len(elements) > 0:
            return elements[0]

    def get_body_element(self):
        """
        Gets the first ``<body>`` element.

        Returns
        -------
        Element

        None
            If the ``<body>`` element was not found.

        """
        elements = self.get_elements("body")
        if len(elements) > 0:
            return elements[0]

    @_PageSignal()
    def add_function(self, func):
        """
        Adds a function to the `Page`. This function can be called from an HTML element.

        Examples
        --------

        Consider an HTML code that contains the following:

        >>> html_with_function = '<html><button onclick="printValue()">Print</button></html>'

        Notice that the ``<button>`` element contains the attribute ``onclick`` calling a Python function `printValue`.
        However, this function is not yet added to the HTML page. To add the function, define it in Python and add it
        to a `Page`:

        >>> def printValue():
        ...     print("value")
        >>> page = Page(html_str=html_with_function)
        >>> page.add_function(printValue)

        Now create a `Website` and add the `Page` to it:

        >>> from toui import Website
        >>> app = Website()
        >>> app.add_pages(page)

        Parameters
        ----------
        function
            Function to be added to the page.

        Returns
        -------
        None

        """
        if self._app:
            self._app._add_function(func)
        old_functions = copy(self._functions)
        self._functions[func.__name__] = func
        if func.__name__ in old_functions:
            return ""
        script_element = Element("script")
        script_element.set_content(custom_func(func.__name__))
        self.get_elements(tag_name="html")[0].add_content(script_element)
        return func.__name__

    @property
    def view_func(self):
        return self._view_func

    @view_func.setter
    def view_func(self, func):
        self._view_func = lambda: func(self)

    def _add_script(self, template_type="web"):
        script_tag = Element("script")
        script_content = get_script(template_type)
        script_tag.set_content(script_content)
        self.get_elements(tag_name="html")[0].add_content(script_tag)

    def _create_window(self, name, api, assets_folder):
        with _TempHTML(directory=assets_folder, html=self.to_str(),
                       win_kwargs={"title": name, "js_api": api}) as temp_html:
            self.window = temp_html.win
            time.sleep(1)
        api.window = self.window
        return self.window

    def _create_first_window(self, name, api, assets_folder):
        self.window = webview.create_window(title=name, js_api=api)
        def func():
            with _TempHTML(directory=assets_folder, html=self.to_str(),
                           win=self.window) as temp_html:
                time.sleep(1)
        api.window = self.window
        return func

    def _evaluate_js(self, func, kwargs):
        codejs = f"""
        var kwargs = JSON.parse(\'{json.dumps(kwargs)}\')
        {func}(kwargs)
        """
        debug("EVALUATE: " + codejs)
        out = self.window.evaluate_js(codejs)

    @_PageSignal()
    def _open_another_page(self, url):
        return


class _TempHTML:

    def __init__(self, directory, win=None, html="", win_args=(), win_kwargs=None):
        self.directory = directory
        self.win = win
        self.html = html
        self.win_args = win_args
        if win_kwargs is None:
            win_kwargs = {}
        self.win_kwargs = win_kwargs

    def __enter__(self):
        i = ""
        name = f"~toui{i}.html"
        while os.path.exists(f"{self.directory}/{name}"):
            if i == "":
                i = 1
            else:
                i += 1
            name = f"~toui{i}.html"
        self.file = f"{self.directory}/{name}"
        file = open(self.file, "w")
        if self.win:
            self.win.load_url(self.file)
        else:
            self.win = webview.create_window(*self.win_args, **self.win_kwargs, url=self.file)
        file.write(self.html)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.remove(self.file)


if __name__ == "__main__":
    import doctest
    doctest.testmod()


"""
A module that creates instructions "signals" to allow communicating with JavaScript.
"""
import inspect
import json
from toui._helpers import debug
from copy import copy
from functools import wraps


class Signal:

    included_private_methods = ["_open_another_page"]
    no_return_functions = []

    def __init__(self):
        self.ws = None

    def __call__(decorator, func):
        decorator._func = func
        @wraps(func)
        def new_func(self, *args, **kwargs):
            decorator.object = self
            if self._signal_mode:
                original_copy = copy(self)
            value = decorator._func(self, *args, **kwargs)
            if self._signal_mode:
                if self.__class__.__name__ == 'Element':
                    decorator.ws = self._parent_page.__dict__.get("_ws")
                    decorator.evaluate_js = self._parent_page._evaluate_js
                elif self.__class__.__name__ == "Page":
                    decorator.ws = self.__dict__.get("_ws")
                    decorator.evaluate_js = self._evaluate_js
                kwargs = inspect.signature(decorator._func).bind(self, *args, **kwargs)
                kwargs.apply_defaults()
                kwargs = kwargs.arguments
                kwargs['return_value'] = value
                kwargs['object'] = kwargs['self']
                kwargs['original_copy'] = original_copy
                del kwargs['self']
                decorator._call_method(decorator._func, **kwargs)
            if func.__name__ in decorator.no_return_functions:
                return
            return value
        return new_func

    def _call_method(self, func, **kwargs):
        for method_name, method in inspect.getmembers(self, inspect.isfunction):
            if method_name == func.__name__ and (not method_name.startswith("_") or
                                                 method_name in self.included_private_methods):
                signal = method(**kwargs)
                if signal:
                    return self._send(signal)
                else:
                    return

    def _send(self, signal):
        debug(f"SENT: {signal}")
        if self.ws:
            self.ws.send(json.dumps(signal))
        else:
            self.evaluate_js(func=signal['func'], kwargs=signal['kwargs'])



import setuptools
from toui import __version__

with open("README.md", "r") as file:
    long_description = file.read()

name = "ToUI"
version = __version__
author = "Mubarak Almehairbi"
description = "Creates user interfaces (websites and desktop apps) from HTML easily"
package_name = "toui"
requirements = []
with open(f"requirements.txt", "rt") as file:
    for pkg in file.read().splitlines():
        pkg_name = pkg.split("==")[0]
        pkg_version = pkg.split("==")[1]
        pkg_major_version = pkg_version.split(".")[0]
        req = f"{pkg_name}>={pkg_version},<{int(pkg_major_version)+1}"
        requirements.append(req)


setuptools.setup(
    name=package_name,
    version=version,
    author=author,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    py_modules=[package_name],
    install_requires=requirements
)


"""
JavaScript code that is added to HTML files.
"""
web_template = f"""
var conn = "wss"
if (location.protocol == "http:") {{
    conn = "ws"
}}
const socket = new WebSocket(conn + '://' + location.host + '/toui-communicate')
socket.addEventListener('message', ev => {{
    _findAndExecute(ev.data)
  }});
const urlPath = location.pathname
function _toPy(...arguments) {{
    var func = arguments.shift()
    var json = {{func:func,
                args: arguments,
                url: urlPath}}
    json['html'] = _getDoc()
    var jsonstring = JSON.stringify(json)
    if (socket.readyState === 0) {{
        socket.onopen = function() {{
            socket.send(jsonstring);
        }}
    }} else {{
        socket.send(jsonstring)
    }}
}}

"""
    
    
desktop_template = f"""
function _toPy(...arguments) {{
    var func = arguments.shift()
    var json = {{func:func,
                args: arguments}}
    json['html'] = _getDoc()
    var jsonstring = JSON.stringify(json)
    pywebview.api.communicate(jsonstring)
}}
"""


common_template = f"""
function _getDoc() {{
    var input_elements = document.getElementsByTagName("input")
    for (var input_element of input_elements) {{
        input_element.setAttribute("value", input_element.value)
        input_element.setAttribute("checked", input_element.checked)
    }}
    var serializer = new XMLSerializer()
    const doc = serializer.serializeToString(document)
    return doc
    }}

function _setDoc(kwargs) {{
    document.open()
    document.write(kwargs['doc'])
    document.close()
    }}
    
function _getElement(kwargs) {{
    var number = 0
    if ('number' in kwargs) {{
        var number = parseInt(kwargs['number'])
    }}
    if ('parent' in kwargs) {{
        const parent = _getElement(kwargs['parent'])
        const element = parent.querySelectorAll(kwargs['selector'])[number]
        return element
    }} else {{
        const elements = document.querySelectorAll(kwargs['selector'])
        const element = elements[number]
        return element
    }}
}}

function _replaceElement(kwargs) {{
    var old_element = _getElement(kwargs['selector'])
    old_element.outerHTML = kwargs['element']
    }}

function _replaceElements(kwargs) {{
    var elements = document.querySelectorAll(kwargs['selectors'])
    var pyelements = kwargs['elements']
    for (i = 0; i < elements.length; i++) {{
        elements[i].outerHTML = pyelements[i];
        }}
    }}
    
function _setAttr(kwargs) {{
    var element = _getElement(kwargs['selector'])
    element.setAttribute(kwargs['name'], kwargs['value'])
    if (kwargs['name'] == 'value') {{
        element.value = kwargs['value']
    }}
    if (kwargs['name'] == 'checked') {{
        element.checked = kwargs['value']
    }}
}}

function _setContent(kwargs) {{
    var element = _getElement(kwargs['selector'])
    element.innerHTML = kwargs['content']
}}

function _addContent(kwargs) {{
    var element = _getElement(kwargs['selector'])
    element.insertAdjacentHTML("beforeend" ,kwargs['content'])
}}

function _addScript(kwargs) {{
    const script = kwargs['script']
    var new_element = document.createElement('script')
    new_element.innerHTML = script
    document.body.appendChild(new_element)
    }}
    
function _goTo(kwargs) {{
    const url = kwargs['url']
    location.href = location.origin + url
}}

function _resizeEmbed(element) {{
    element.height = 0
    element.width = 0
    element.style.height = element.contentWindow.document.body.scrollHeight + 1 + 'px'
    element.style.width = element.contentWindow.document.body.scrollWidth + 1 + 'px'
}}

function _findAndExecute(jsonString) {{
    var instructions = JSON.parse(jsonString)
    var func = instructions['func']
    var kwargs = instructions['kwargs']
    if (func == "_replaceElement") {{
        _replaceElement(kwargs)
    }}
    if (func == "_replaceElements") {{
        _replaceElements(kwargs)
    }}
    if (func == "_setAttr") {{
        _setAttr(kwargs)
    }}
    if (func == "_setContent") {{
        _setContent(kwargs)
    }}
    if (func == "_addContent") {{
        _addContent(kwargs)
    }}
    if (func == "_setDoc") {{
        _setDoc(kwargs)
    }}
    if (func == "_addScript") {{
        _addScript(kwargs)
    }}
    if (func == "_goTo") {{
        _goTo(kwargs)
    }}
}}"""

def custom_func(name):
    text = f"""
    function {name}(...args) {{
        _toPy('{name}', ...args)
    }}
    """
    return text


def get_script(app_type='web'):
    if app_type == "desktop":
        template = desktop_template
    else:
        template = web_template
    script = template + common_template
    return script
